//------------------------------------------------------
// module  : Denoising.cc
// auteur  : Mignotte Max
// date    : 
// version : 1.0
// langage : C++
// note    :
//------------------------------------------------------
//  

//------------------------------------------------
// FICHIERS INCLUS -------------------------------
//------------------------------------------------
#include "Denoising.h"


//------------------------------------------------
// DEFINITIONS -----------------------------------
//------------------------------------------------
#define NAME_IMG       "Image"
#define NAME_DEGRAD    "ImgDegrad"
#define NAME_SEGS      "ImgSegs"
#define NAME_REGS      "ImgRegions"
#define NAME_REST      "ImgRest" 

#define NCHAR 200

//------------------------------------------------
// CONSTANTES ------------------------------------
//------------------------------------------------
const int   METHOD=2;

const int ZOOM=1;
const int QUIT=0;
const int SAVE=0;

//------------------------------------------------
// VARIABLES GLOBALES ----------------------------
//------------------------------------------------
char Name_img[NCHAR];
char Name_imgdeg[NCHAR];

int   method;
int   nbiter;

int    nbsegs;
float  variance_noise;
float  varclusters;
float  hyper;
float* hyperpar;
int    nbclass;
int    seed;
int    sizemax;
int    nbclassinf,nbclasssup;
float  betainf,betasup;

int   zoom;

int   flag_save;
int   flag_quit;

//----------------------------------------
//--- Windows ----------------------------
//----------------------------------------
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>

Display   *display;
int	  screen_num;
int 	  depth;
Window	  root;
Visual*	  visual;
GC	  gc;

/************************************************************************/
/* OPEN_DISPLAY()							*/
/************************************************************************/
int open_display()
{
  if ((display=XOpenDisplay(NULL))==NULL)
   { printf("Connection impossible\n");
     return(-1); }

  else
   { screen_num=DefaultScreen(display);
     visual=DefaultVisual(display,screen_num);
     depth=DefaultDepth(display,screen_num);
     root=RootWindow(display,screen_num);
     return 0; }
}

/************************************************************************/
/* FABRIQUE_WINDOW()							*/
/* Cette fonction cre une fenetre X et l'affiche  l'cran.	        */
/************************************************************************/
Window fabrique_window(char *nom_fen,int x,int y,int width,int height,int zoom)
{
  Window                 win;
  XSizeHints      size_hints;
  XWMHints          wm_hints;
  XClassHint     class_hints;
  XTextProperty  windowName, iconName;

  char *name=nom_fen;

  if(zoom<0) { width/=-zoom; height/=-zoom; }
  if(zoom>0) { width*=zoom;  height*=zoom;  }

  win=XCreateSimpleWindow(display,root,x,y,width,height,1,0,255);

  size_hints.flags=PPosition|PSize|PMinSize;
  size_hints.min_width=width;
  size_hints.min_height=height;

  XStringListToTextProperty(&name,1,&windowName);
  XStringListToTextProperty(&name,1,&iconName);
  wm_hints.initial_state=NormalState;
  wm_hints.input=True;
  wm_hints.flags=StateHint|InputHint;
  class_hints.res_name=nom_fen;
  class_hints.res_class=nom_fen;

  XSetWMProperties(display,win,&windowName,&iconName,
                   NULL,0,&size_hints,&wm_hints,&class_hints);

  gc=XCreateGC(display,win,0,NULL);

  XSelectInput(display,win,ExposureMask|KeyPressMask|ButtonPressMask| 
               ButtonReleaseMask|ButtonMotionMask|PointerMotionHintMask| 
               StructureNotifyMask);

  XMapWindow(display,win);
  return(win);
}

/****************************************************************************/
/* CREE_XIMAGE2()							    */
/* Cre une XImage  partir d'un tableau de float                           */
/* L'image peut subir un zoom.						    */
/****************************************************************************/
XImage* cree_Ximage2(float** mat,int z,int length,int width)
{
  int lgth,wdth,lig,col,zoom_col,zoom_lig;
  float somme;
  unsigned char	 pix;
  unsigned char* dat;
  XImage* imageX;

  /*Zoom positiv*/
  /*------------*/
  if (z>0)
  {
   lgth=length*z;
   wdth=width*z;

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<lgth;lig=lig+z) for(col=0;col<wdth;col=col+z)
   { 
    pix=(unsigned char)mat[lig/z][col/z];
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
      { 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+0)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+1)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+2)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+3)]=pix; 
       }
    }
  } /*--------------------------------------------------------*/

  /*Zoom negatifv*/
  /*------------*/
  else
  {
   z=-z;
   lgth=(length/z);
   wdth=(width/z);

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<(lgth*z);lig=lig+z) for(col=0;col<(wdth*z);col=col+z)
   {  
    somme=0.0;
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
     somme+=mat[lig+zoom_lig][col+zoom_col];
           
     somme/=(z*z);    
     dat[((lig/z)*wdth*4)+((4*(col/z))+0)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+1)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+2)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+3)]=(unsigned char)somme; 
   }
  } /*--------------------------------------------------------*/

  imageX=XCreateImage(display,visual,depth,ZPixmap,0,(char*)dat,wdth,lgth,16,wdth*4);
  return (imageX);
}

/****************************************************************************/
/* CREE_XIMAGE4()							    */
/* Cre une XImage  partir d'un tableau d'entier                           */
/* L'image peut subir un zoom.						    */
/****************************************************************************/
XImage* cree_Ximage4(int** mat,int z,int length,int width)
{
  int lgth,wdth,lig,col,zoom_col,zoom_lig;
  float somme;
  unsigned char	 pix;
  unsigned char* dat;
  XImage* imageX;

  /*Zoom positif*/
  /*------------*/
  if (z>0)
  {
   lgth=length*z;
   wdth=width*z;

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<lgth;lig=lig+z) for(col=0;col<wdth;col=col+z)
   { 
    pix=(unsigned char)mat[lig/z][col/z];
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
      { 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+0)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+1)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+2)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+3)]=pix; 
       }
    }
  } /*--------------------------------------------------------*/

  /*Zoom negatif*/
  /*------------*/
  else
  {
   z=-z;
   lgth=(length/z);
   wdth=(width/z);

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<(lgth*z);lig=lig+z) for(col=0;col<(wdth*z);col=col+z)
   {  
    somme=0.0;
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
     somme+=mat[lig+zoom_lig][col+zoom_col];
           
     somme/=(z*z);    
     dat[((lig/z)*wdth*4)+((4*(col/z))+0)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+1)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+2)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+3)]=(unsigned char)somme; 
   }
  } /*--------------------------------------------------------*/

  imageX=XCreateImage(display,visual,depth,ZPixmap,0,(char*)dat,wdth,lgth,16,wdth*4);
  return (imageX);
}

/****************************************************************************/
/* CREE_XIMAGECOUL()							    */
/* Cre une XImage  partir d'un tableau 3 d d'entier                       */
/* L'image peut subir un zoom.						    */
/****************************************************************************/
XImage* cree_XimageCoul(int*** matRVB,int z,int length,int width)
{
  int i;
  int lgth,wdth,lig,col,zoom_col,zoom_lig;
  float somme;
  float sum[3];
  unsigned char	 pixR,pixV,pixB,pixN;
  unsigned char* dat;
  XImage* imageX;

  /*Zoom positif*/
  /*------------*/
  if (z>0)
  {
   lgth=length*z;
   wdth=width*z;

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<lgth;lig=lig+z) for(col=0;col<wdth;col=col+z)
   { 
    pixR=(unsigned char)matRVB[0][lig/z][col/z];
    pixV=(unsigned char)matRVB[1][lig/z][col/z];
    pixB=(unsigned char)matRVB[2][lig/z][col/z];
    somme=(1.0/3.0)*(pixR+pixV+pixB);
    pixN=(unsigned char)somme;

    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
      { 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+0)]=pixB; 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+1)]=pixV; 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+2)]=pixR; 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+3)]=0; 
       }
    }
  } /*--------------------------------------------------------*/

  /*Zoom negatif*/
  /*------------*/
  else
  {
   z=-z;
   lgth=(length/z);
   wdth=(width/z);

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<(lgth*z);lig=lig+z) for(col=0;col<(wdth*z);col=col+z)
   {  
    sum[0]=sum[1]=sum[2]=0.0;
    
    for(i=0;i<3;i++)
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
     sum[i]+=matRVB[i][lig+zoom_lig][col+zoom_col];
       
    for(i=0;i<3;i++)  sum[i]/=(z*z); 

     dat[((lig/z)*wdth*4)+((4*(col/z))+0)]=(unsigned char)sum[1];
     dat[((lig/z)*wdth*4)+((4*(col/z))+1)]=(unsigned char)sum[1];
     dat[((lig/z)*wdth*4)+((4*(col/z))+2)]=(unsigned char)sum[1];
     dat[((lig/z)*wdth*4)+((4*(col/z))+3)]=(unsigned char)sum[1]; 
   }
  } /*--------------------------------------------------------*/

  imageX=XCreateImage(display,visual,depth,ZPixmap,0,(char*)dat,wdth,lgth,16,wdth*4);
  return (imageX);
}

//----End Windows---------------------------------------------------------
//------------------------------------------------------------------------

//---------------------------------------------------------
// Enregistre les arguments de la ligne de commande      
//---------------------------------------------------------
void read_arguments(int argc,char** argv)
{
 int i;

   //Options
  if (argc<2) 
    { printf("\n Usage Denoising [Img]"); 
      printf("\n Options are : value by default indicated in ()");
      printf("\n");
      printf("\n            -------------");
      printf("\n            -m # (%d)",METHOD);
      printf("\n");
      printf("\n            -m 2   DenoiseMigNSegs  NbClasses Aleatoires  [2]");
      printf("\n            --------------");
      printf("\n            -v  Variance Bruit (%d)",VARIANCE_NOISE);
      printf("\n            -------------");
      printf("\n            -z zoom (%d)",ZOOM);
      printf("\n            -q quit (Sans)"); 
      printf("\n            -s save (Sans)"); 
      printf("\n            -------------");
      printf("\n");
      printf("\n    Ex: Denoising lena512.pgm -v 900"); 
      printf("\n\n\n");
      exit(-1); }

      //Chargement Fichier
      if (argc>1) strcpy(Name_img,argv[1]);    

   //Boucle
   for(i=2;i<argc;i++)
     {   
          switch(argv[i][1])
          {
           case 'm': method=atoi(argv[++i]);               break;
	   case 'v': variance_noise=atof(argv[++i]);       break;
           case 's': flag_save=1;                          break; 
           case 'z': zoom=atoi(argv[++i]);                 break;
	   case 'q': flag_quit=1;                          break;           
          }    
     } 
 }

//----------------------------------------------------------
//----------------------------------------------------------
// PROGRAMME PRINCIPAL -------------------------------------
//----------------------------------------------------------
//----------------------------------------------------------
int main(int argc,char** argv)
{
 int i;
 int   flag;
 int length,width;

 //Pour Xwindow
 //------------
 XEvent ev;
 Window win_ppicture_img;
 Window win_ppicture_deg;
 Window win_ppicture_segs;
 Window win_ppicture_regs;
 Window win_ppicture_rest;

 XImage *x_ppicture_img;
 XImage *x_ppicture_deg;
 XImage *x_ppicture_segs;
 XImage *x_ppicture_regs; 
 XImage *x_ppicture_rest;

 char nomfen_ppicture_img[NCHAR];
 char nomfen_ppicture_deg[NCHAR];
 char nomfen_ppicture_segs[NCHAR];
 char nomfen_ppicture_regs[NCHAR];
 char nomfen_ppicture_rest[NCHAR];
 
 //Initialisation
 //--------------
 method=METHOD;
 variance_noise=VARIANCE_NOISE;
 
 nbsegs=NB_SEGMENTATIONS;
 varclusters=VARCLUSTERS;
 nbiter=ITERATION_ECI;
 nbclass=NBCLASS;
 seed=SEED;

 zoom=ZOOM;
 flag_save=SAVE;
 flag_quit=QUIT;
 flag=0;

 hyperpar=fmatrix_allocate_1d(NB_HYPERPAR);
 for(i=0;i<NB_HYPERPAR;i++)  hyperpar[i]=VALUE_HYPERPAR;
 hyper=VALUE_HYPERPAR;


 //Lis Arguments
 //-------------
 read_arguments(argc,argv);

 //Initialisation Parametres internes
 //----------------------------------
 sizemax=(int)sqrt(variance_noise);

 for(i=0;i<NB_HYPERPAR;i++)  hyperpar[i]=MAX(0.75,sqrt(variance_noise)/20.0);
 hyper=MAX(0.75,sqrt(variance_noise)/20.0); 

 nbclassinf=3;
 nbclasssup=(int)((256/(2*sqrt(variance_noise+0.00000001)))+1); 
 if (nbclasssup<6)  nbclasssup=6; 
 if (nbclasssup>13) nbclasssup=13; 


 //Affiche options
 //---------------
 printf("\n\n Info: ");
 printf("\n -------------------------------------------");
 printf("\n Ouverture de l'image :   <%s>",Name_img);
 printf("\n Variance noise           [%.0f]",variance_noise); 
 printf("\n -----");
 printf("\n Methode de Denoising     [%d]",method);
 printf("\n -------------------------------------------");
 printf("\n Nb Iteration:                  [%d]",nbiter);
 printf("\n Nb Simulations:                [%d]",nbsegs);
 printf("\n Nb Classes:                    [%d]",nbclass);
 printf("\n Value Hyper: [%.2f][%.2f][%.2f][%.2f]",hyperpar[0],hyperpar[1],hyperpar[2],hyperpar[3]);
 printf("\n Value Seed:                    [%d]",seed); 
 printf("\n Variance Clusters:             [%.3f]",varclusters);
 printf("\n Size Max:                      [%d]",sizemax);
 printf("\n Nb Classes Inf/Sup:            [%d::%d::%d]",nbclassinf,nbclass,nbclasssup);
 printf("\n Beta Inf/Sup:                  [%.2f::%.2f::%.2f]",betainf,hyper,betasup);
 printf("\n -------------------------------------------"); 
 printf("\n Zoom: [%d]",zoom); 
 if (flag_save)   printf("\n -> Avec Sauvegarde");
 if (flag_quit)   printf("\n -> Avec Quit");

//-------------------------------
//---- Chargement des Images -------------------------------
//-------------------------------

 //Recuperation Images & Info
  //--------------------------
  GetLengthWidth(Name_img,&length,&width);
  int** datai=imatrix_allocate_2d(length,width);
  load_image_pgm(Name_img,datai,length,width);
 
  //Information 
  printf("\n Taille de l'image: [%d - %d]",length,width);
  fflush(stdout); 


  //------------------------------------------
  //-------- Algorithme DENOISE --------------
  //------------------------------------------
  Denoise denois(datai,length,width);

  denois.option(flag_save,Name_img);
  denois.optionSeg(nbclass,hyper,seed,varclusters,sizemax,nbiter,nbclassinf,nbclasssup,betainf,betasup);
  denois.optionNSeg(nbsegs);

  denois.DenoisingChoice(method,variance_noise);

//---------------------------------------------------------------
//---------------- visu sous XWINDOW ----------------------------
//---------------------------------------------------------------
 if (!flag_quit)
 {
 //ouverture session graphique
 if (open_display()<0) printf(" Impossible d'ouvrir une session graphique");

 //commentaire dans les photos affichees
 sprintf(nomfen_ppicture_img,"%s : %s",NAME_IMG,Name_img);
 sprintf(nomfen_ppicture_deg,"%s : %s",NAME_DEGRAD,Name_img);
 sprintf(nomfen_ppicture_segs,"%s : %s",NAME_SEGS,Name_img);
 sprintf(nomfen_ppicture_regs,"%s : %s",NAME_REGS,Name_img);
 sprintf(nomfen_ppicture_rest,"%s : %s",NAME_REST,Name_img);
 
 //creation des fenetres
 win_ppicture_img=fabrique_window(nomfen_ppicture_img,10,10,width,length,zoom);
 win_ppicture_deg=fabrique_window(nomfen_ppicture_deg,10,10,width,length,zoom);
 win_ppicture_segs=fabrique_window(nomfen_ppicture_segs,10,10,width*nbsegs,length,zoom);
 win_ppicture_regs=fabrique_window(nomfen_ppicture_regs,10,10,width*nbsegs,length,zoom); 
 win_ppicture_rest=fabrique_window(nomfen_ppicture_rest,10,10,width,length,zoom);

 //creation des Ximages
 x_ppicture_img=cree_Ximage4(datai,zoom,length,width);
 x_ppicture_deg=cree_Ximage2(denois.mat_deg,zoom,denois.length,denois.width);
 x_ppicture_segs=cree_Ximage4(denois.win_tabmat_segneb,zoom,denois.length,(denois.width)*nbsegs);
 x_ppicture_regs=cree_XimageCoul(denois.win_tabmat_regcoul,zoom,denois.length,(denois.width)*nbsegs);
 x_ppicture_rest=cree_Ximage2(denois.mat_rest,zoom,denois.length,denois.width);

 printf("\n\n Pour quitter,appuyer sur la barre d'espace\n");
 fflush(stdout);

 //boucle d'evenements
  for(;;)
     {
      XNextEvent(display,&ev);
       switch(ev.type)
        {
         case Expose:
         XPutImage(display,win_ppicture_img,gc,x_ppicture_img,0,0,0,0,
                   x_ppicture_img->width,x_ppicture_img->height);
         XPutImage(display,win_ppicture_deg,gc,x_ppicture_deg,0,0,0,0,
                   x_ppicture_deg->width,x_ppicture_deg->height);
         XPutImage(display,win_ppicture_segs,gc,x_ppicture_segs,0,0,0,0,
                   x_ppicture_segs->width,x_ppicture_segs->height);
         XPutImage(display,win_ppicture_regs,gc,x_ppicture_regs,0,0,0,0,
                   x_ppicture_regs->width,x_ppicture_regs->height);
         XPutImage(display,win_ppicture_rest,gc,x_ppicture_rest,0,0,0,0,
                   x_ppicture_rest->width,x_ppicture_rest->height);
         break;

         case KeyPress:
         XDestroyImage(x_ppicture_img);
         XDestroyImage(x_ppicture_deg);
         XDestroyImage(x_ppicture_segs);
         XDestroyImage(x_ppicture_regs);
         XDestroyImage(x_ppicture_rest);

         XFreeGC(display,gc);
         XCloseDisplay(display);
         flag=1;
         break;
         }
   if (flag==1) break;
   }
 } 
       
//--------------- fin de la session graphique --------------     
//----------------------------------------------------------

   //Liberation Memoire
   //------------------
   if (datai) free_imatrix_2d(datai);

   //retour sans probleme
   //--------------------
   printf("\n C'est fini... \n\n\n");
   exit(-1);
   return 0;
 }
 


