//---------------------------------------------------
// module  : Denoising.h                         
// auteur  : Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

#ifndef DENOISING_H
#define DENOISING_H

//------------------------------------------------
// FICHIERS INCLUS -------------------------------
//------------------------------------------------
#include "FunctionDen.h"
#include "GibbsSampler.h"
#include "Eci.h"
#include "kmean.h"
#include "Region.h" 

//------------------------------------------------
// DEFINITIONS -----------------------------------
//------------------------------------------------

#define VARIANCE_NOISE    400   //(by default)

#define ITERATION_ECI     6
#define NBCLASS           6
#define SEED              1
#define NB_HYPERPAR       4
#define NB_PAR            3
#define VALUE_HYPERPAR    1.0
#define VARCLUSTERS       0.0250 
#define NB_SEGMENTATIONS  60
#define SIZE_MAX          15
#define NB_CLASSES_MAX      30
#define NB_CANAUX_COULEURS   3
#define CHGT               300

#define NB_GREY_LEVELS 256
#define DEBUG 0

//------------------------------------------------
// CLASSE DENOISE --------------------------------
//------------------------------------------------
class Denoise
{
  //--Donnees--
  //-----------
  public:

  //Matrices
  int**   matx_img;
  int**   matx_tmp1;
 
  float** mat_img;
  float** mat_deg;
  float** mat_seg;  
  float** mat_rest;
 
  float** mat_tmp4; 
  int***   mat_regcoul; 
  

  //Windows
  int***   win_tabmat_regcoul;
  int**    win_tabmat_segneb;

  //Donnee
  int length,width,size_image;

  //Pour Multiseg
  int****   tabmat_regcoul;   //Pour visu MultiSeg-Regions
  float***  mat_const;        //Pour RegionBasedConstraints
  int***    tabmat_segneb;    //Pour visu MultiSeg
  
  //Methode-Parametre Restoration
  int   method;
  float variance_noise;
 
  //Segmentation
  int      nbiter;
  int      seed;
  int      nbclass;
  int      nbclasscur,nbclassinf,nbclasssup;
  float    betacur,betainf,betasup;
  float    value_hyperpar;
  float*   hyperpar;
  float**  pparseg;
  float**  tabpts;
  int      nbsegs;
  int      simulations;
  float    varclusters;
  int      sizemax;

  //Set-Regions
  int   NbRegions;

  //NRJ & Mesure
  float isnr_restor;
  float mmse_degrad;
  float mmse_restor;
  float snr_degrad;
  float snr_true_degrad;
  float snr_restor;
  float snr_true_restor;
  float psnr_degrad;
  float psnr_restor;

  //Stats
  int* HistoClasses;
  int* HistoBeta;

  //Flag
  //int  flag_eci;
  int  flag_save;
  int  flagR_silent;
  int  flag_ssilent;

  //Utile
  char Name_Rest[300];
  char Name_Sauve[300];
  char Name_Img[300];
  char Par_Result[300];

  //-----Objets-----
  //----------------
  kmean*              km;
  Eci*               eci;
  SetRegion*      setreg;
  SetRegion**  tabsetreg;
  
  //--Methodes--
  //------------
  public:
  Denoise(int**,int,int);
  ~Denoise();

  //Reset
  void Reset();

  //Options
  void option(int,char*);
  void optionSeg(int,float,int,float,int,int,int,int,float,float);
  void optionNSeg(int);
  
  //Choix Methode
  void  DenoisingChoice(int,float);

  //Methodes Denoisings
  void  DenoiseMigNSegsNbClAleat();
  
  //-------------------------------------------------
  // Fonctions
  //-------------------------------------------------  
  //Degradation
  void Degradation(float**,float**,float);

  //Segmentations
  void  SeveralSegmentationsNbClAleat(float**);
  void  AutomaticEciSimulation(float**,int);
  void  EstimateSetRegions(SetRegion*);

  //Windows
  void  ComputeMatSegMoy(float**,float**);
  void  ComputeWindows();
  
  //Mesure
  float computeISNR(float**,float**,float**);
  float computeMMSE(float**,float**,int);
  float computePSNR(float**,float**,int);
  float computeSNR(float**,float,int);
  float computeSNR(float**,float**,int);

  //Utile
  void   saveImagePgm(char*,char*,float**,int,int);
  void   Colorization(float**,SetRegion*);
  void   FillResRestoration();
  void   CopyVect(float**,float**);
  void   Visu_screen(int,float**);
  void   DiffMat(float**,float**,float**,int);
};

#endif 
