//---------------------------------------------------
// module  : f_Denoising.cc                       
// auteur  : Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "Denoising.h"

//------------------------------------------------
// DEFINITIONS -----------------------------------
//------------------------------------------------
//------------------------------------------------
// METHODES DE LA CLASSE DENOISE  -----------------
//------------------------------------------------
//----------------------------------------------------------
// Constructeur de la classe Denoise
//----------------------------------------------------------
Denoise::Denoise(int** imgt,int lgth,int wdth)           
 {
  int i,j;

  //Initialisation
  length=lgth;
  width=wdth;

  if (length!=width) { printf("\n Probleme Image non carree"); exit(-1); }
  size_image=length;

  //Allocation memoire matrice int**
  matx_img=imatrix_allocate_2d(length,width);  
  matx_tmp1=imatrix_allocate_2d(length,width);  
 
  //Allocation memoire matrice float**
  mat_img=fmatrix_allocate_2d(length,width);
  mat_deg=fmatrix_allocate_2d(length,width);
  mat_seg=fmatrix_allocate_2d(length,width);
  mat_rest=fmatrix_allocate_2d(length,width);
 
  mat_tmp4=fmatrix_allocate_2d(length,width); 

  //Allocation memoire matrice int***
  mat_regcoul=imatrix_allocate_3d(NB_CANAUX_COULEURS,length,width);

  //Allocation memoire pour seg
  hyperpar=fmatrix_allocate_1d(NB_HYPERPAR);
  pparseg=fmatrix_allocate_2d((NB_CLASSES_MAX)*2,NB_PAR); 
  tabpts=fmatrix_allocate_2d(width*length,1);
  //tabpts2=fmatrix_allocate_2d(width*length,3);

  //Initialisation Matrix
  for(i=0;i<length;i++) for(j=0;j<width;j++) 
     { mat_img[i][j]=imgt[i][j];
       matx_img[i][j]=imgt[i][j]; }

  //Initialisation Parametre
  nbclasscur=0;
  nbclassinf=0;
  nbclasssup=0;
  betainf=0.0;
  betasup=0.0;

  //Initialisation FLAG
  flagR_silent=1;  //Set to one for information
  flag_ssilent=1;  //Set to one for information
 }

//----------------------------------------------------------
// Destructeur de la classe Denoise
//----------------------------------------------------------
Denoise::~Denoise()
{ 
  //Libere Memoire Matrice int**
  if (matx_img) free_imatrix_2d(matx_img);
  if (matx_tmp1) free_imatrix_2d(matx_tmp1);
  
  //Libere Memoire Matrice float**
  if (mat_img)  free_fmatrix_2d(mat_img);
  if (mat_deg)  free_fmatrix_2d(mat_deg); 
  if (mat_seg)  free_fmatrix_2d(mat_seg);
  if (mat_rest) free_fmatrix_2d(mat_rest);

  if (mat_tmp4)  free_fmatrix_2d(mat_tmp4); 
  
  //Libere Memoire Matrice int***
  if (mat_regcoul)    free_imatrix_3d(mat_regcoul,TROIS);

  //Libere Matrice Segmentations
  if (mat_const)      free_fmatrix_3d(mat_const,nbsegs);
  if (tabmat_regcoul) free_imatrix_4d(tabmat_regcoul,nbsegs,TROIS); 

  if (win_tabmat_regcoul) free_imatrix_3d(win_tabmat_regcoul,NB_CANAUX_COULEURS);
  if (win_tabmat_segneb)  free_imatrix_2d(win_tabmat_segneb);

  //Libere Memoire pour seg
  if (hyperpar) free_fmatrix_1d(hyperpar);
  if (pparseg)  free_fmatrix_2d(pparseg);
  if (tabpts)   free_fmatrix_2d(tabpts);

  //Libere Objet
  for(int k=0;k<nbsegs;k++) if (tabsetreg[k]) delete(tabsetreg[k]);
  if (tabsetreg) delete[](tabsetreg);
}
//----------------------------------------------------------
// Reset
//----------------------------------------------------------
void Denoise::Reset()
{ 
 int k;

 //Libere Objet
 for(k=0;k<nbsegs;k++) if (tabsetreg[k]) delete(tabsetreg[k]);
}


//-----------//
//--OPTIONS--//
//-----------//
//----------------------------------------------------------
// Options                  
//----------------------------------------------------------                
void Denoise::option(int sv,char* ch)
{

 //Enregistrement
 flag_save=sv;
 strcpy(Name_Img,ch);
 
 //Info
 if ((1)&&(!flag_ssilent))
    { printf("\n Sauvegarde             [%d]",flag_save);
      printf("\n Nom Image              [%s]",Name_Img); }
}

//----------------------------------------------------------
// Options pour la segmentation               
//----------------------------------------------------------                
void Denoise::optionSeg(int nbc,float hyp,int sd,float vcl,int szm,int nbit,int nbci,int nbcs,float binf,float bsup)
{
 int i;

 //Enregistrement
 value_hyperpar=hyp;
 for(i=0;i<NB_HYPERPAR;i++) hyperpar[i]=value_hyperpar;
 nbclass=nbc;
 seed=sd;
 varclusters=vcl;
 sizemax=szm;
 nbiter=nbit;
 nbclassinf=nbci;
 nbclasssup=nbcs;
 betainf=binf;
 betasup=bsup; 

 //Info
 if ((1)&&(!flag_ssilent))
    { 
     printf("\n Nb Iters:    [%d]",nbiter);
     printf("\n Nb Classes:  [%d]",nbclass);
     printf("\n Value Hyper: [%.2f][%.2f]",hyperpar[0],hyperpar[1]);
     printf("[%.2f][%.2f]",hyperpar[2],hyperpar[3]);
     printf("\n Value Seed: [%d]",seed); 
     printf("\n Variance Clusters [%.3f]",varclusters);
     printf("\n Size Max [%d]",sizemax);
     printf("\n Nb Classes Inf/Sup [%d -%d- %d]",nbclassinf,nbclass,nbclasssup);
     printf("\n Beta Inf/Sup [%.3f -%.3f- %.3f]",betainf,value_hyperpar,betasup);
    }
}

//----------------------------------------------------------
// Options pour la N segmentations               
//----------------------------------------------------------                
void Denoise::optionNSeg(int nbseg)
{
 //Enregistrement
 nbsegs=nbseg;

 if ((1)&&(!flag_ssilent))
 printf("\n Nb. Segmentations: [%d]",nbsegs);

 //Allocation Memoire
 mat_const=fmatrix_allocate_3d(nbsegs,length,width);
 tabmat_segneb=imatrix_allocate_3d(nbsegs,length,width);
 tabmat_regcoul=imatrix_allocate_4d(nbsegs,NB_CANAUX_COULEURS,length,width); 

 //Windows
 win_tabmat_regcoul=imatrix_allocate_3d(NB_CANAUX_COULEURS,length,width*nbsegs);
 win_tabmat_segneb=imatrix_allocate_2d(length,width*nbsegs);

 //Allocation Memoire  tabsetreg
 tabsetreg=new SetRegion*[nbsegs];
}


//-----------------//
//--CHOIX METHODE--//
//-----------------//
//----------------------------------------------------------
//  Methode d'Initialisation de Denoising                     
//----------------------------------------------------------                
void Denoise::DenoisingChoice(int mtd,float vrnoise)
{
 int k;
 float tmp;

 //Initialisation
 method=mtd;
 variance_noise=vrnoise;
 
 //Information
 if (!flag_ssilent)
    { printf("\n -----");
      printf("\n Methode  [%d]",method);
      printf("\n Varnoise [%.1f]",variance_noise);
      printf("\n Value Seed: [%d]",seed); 
      printf("\n -----");
      printf("\n"); }

 //seed
 for(k=0;k<seed;k++) tmp=(float)rand();

 //Degradation
 CopySquareMatrix(mat_img,mat_deg,size_image);
 Degradation(mat_img,mat_deg,variance_noise);
   
  if (flag_save)
  saveImagePgm("matdeg","",mat_deg,length,width);

  //Calcul PSNR-SNR
  psnr_degrad=computePSNR(mat_img,mat_deg,size_image);
  snr_degrad=computeSNR(mat_img,variance_noise,size_image);
  snr_true_degrad=computeSNR(mat_deg,mat_img,size_image);
  mmse_degrad=computeMMSE(mat_img,mat_deg,size_image);

  printf("\n\n  Degradation Mesure:");
  printf("\n  -------------------");
  printf("\n    >> MMSE : %.3f ",mmse_degrad); 
  printf("\n    >> PSNR : %.3f dB",psnr_degrad);
  printf("\n    >>  SNR : %.3f dB  [%.3f dB]",snr_true_degrad,snr_degrad);
  
 
 //Methode Considr [matx_est] > [mat_rest]
 //-----------------------------------------
 DenoiseMigNSegsNbClAleat();

 //-----------------------------------------

 //ISNR [mat_rest]
 isnr_restor=computeISNR(mat_img,mat_deg,mat_rest);
 mmse_restor=computeMMSE(mat_img,mat_rest,size_image);
 psnr_restor=computePSNR(mat_img,mat_rest,size_image);
 snr_restor=computeSNR(mat_rest,mat_img,size_image);

 printf("\n\n"); 
 printf("\n Resultat Denoising");
 printf("\n ------------------");
 printf("\n MMSE = %.3f",mmse_restor);
 printf("\n ---");
 printf("\n ISNR = %.3f dB",isnr_restor);
 printf("\n PSNR = %.3f dB",psnr_restor);
 printf("\n  SNR = %.3f dB",snr_restor); 
 printf("\n");  

 //Parametre Resultat-Sauvegarde
 strncpy(Name_Sauve,Name_Img,4);
 FillResRestoration();
 if (!flag_ssilent) printf("\n   > Extension Sauvegarde");

 if (!flag_ssilent) printf("\n   > [%s][%s]",Name_Sauve,Par_Result);
 if (flag_ssilent)  printf("\n   > [%s]",Par_Result);

 //Method Noise
 //------------
 DiffMat(mat_tmp4,mat_deg,mat_rest,size_image);
 saveImagePgm("MethodNoise","",mat_tmp4,size_image,size_image); 

 //Sauvegarde
 //if (flag_save) 
 saveImagePgm(Name_Sauve,Par_Result,mat_rest,length,width);
}

//------------------------//
//--METHODE DENOISING-----//
//------------------------//
//----------------------------------------------------------
// DenoiseMigNSegsNbClAleat
//----------------------------------------------------------     
void Denoise::DenoiseMigNSegsNbClAleat()
{
 int k;
 
 //Init & Default
 nbclasscur=(int)nbclass; 
 if (!nbclassinf)
    { nbclassinf=(int)(nbclass-(nbclass*30.0/100.0));
      nbclasssup=(int)(nbclass+(nbclasscur-nbclassinf)); }
 
 HistoClasses=new int[nbclasssup+2];
 for(k=0;k<=nbclasssup+1;k++) HistoClasses[k]=0;

 //Presentation
 if (!flag_ssilent)
    { printf("\n     ----------------------------------"); 
      printf("\n       DenoiseMigNSegsNbClAleat (-m 2) ");
      printf("\n     ----------------------------------");
      printf("\n      >> -m 2  DenoiseMigNSegsNbClAleat <<");
      printf("\n     (> VarianceNoise: %.1f)",variance_noise);
      printf("\n     (> SizeImage: %d)",size_image);
      printf("\n     (> NbClasses: [%d-%d-%d])",nbclassinf,nbclass,nbclasssup);
      printf("\n"); }

 
 // DENOISING - 
 //----------------------------
 SeveralSegmentationsNbClAleat(mat_deg);                              

 //Calcul Moyenne  
 ComputeMatSegMoy(mat_rest,mat_deg);
 ComputeWindows(); 

 //End
 nbclass=nbclasscur;

 //Stats Nb-Classes
 if (!flag_ssilent)
    { fflush(stdout);
      printf("\n\n");
      printf("\n  > NbClasses: [%d-%d-%d]) \n  ",nbclassinf,nbclass,nbclasssup);
      for(k=1;k<=nbclasssup+1;k++) printf("[%d>%d]: ",k,HistoClasses[k]);
      printf("\n\n");  
      fflush(stdout); }
}

//------------------------//
//-----FONCTIONS----------//
//------------------------//
//----------------------------------------------------------
// Degradation        
//----------------------------------------------------------                
void Denoise::Degradation(float** mtin,float** mtout,float vn)
{
  int i,j;
  float varv;
  int nbsup;
  int nbinf;
  float pourcnbsup;
  float pourcnbinf;

  //Presentation
  if (!flag_ssilent)
     { printf("\n");
       printf("\n  Degradation ");
       printf("\n  ----------- "); }
  
 //Ajout BRUIT
 //-----------
 varv=0.0;
 nbsup=0;
 nbinf=0;
 
 add_gaussian_noise(mtout,length,width,vn);

 for(i=0;i<length;i++) for(j=0;j<width;j++)
   { 
    if (mtout[i][j]>255) 
      {   mtout[i][j]=255; 
          nbsup++; }
    if (mtout[i][j]<0)   
      {   mtout[i][j]=0;   
          nbinf++; }

    varv=CARRE(mtout[i][j]-mtin[i][j]);
   }

 pourcnbsup=(100.0*nbsup)/(length*width);
 pourcnbinf=(100.0*nbinf)/(length*width);

 if (!flag_ssilent)
 printf("\n    [%d sups (%.2f PCent) | %d infs (%.2f PCent)]",nbsup,pourcnbsup,nbinf,pourcnbinf);
 
}

//----------------------//
//----SEGMENTATIONS-----//
//----------------------//
//----------------------------------------------------------
//  SeveralSegmentations 
//
//  [mt] > [regcoul] > 
//         
//----------------------------------------------------------     
void Denoise::SeveralSegmentationsNbClAleat(float** mt)
{
 int k;
 char Number[10];

  //Affiche Options
  //---------------
  printf("\n\n  Simulations ECI [%d segmentations]:\n     ",nbsegs);

  if (!flag_ssilent)
     { printf("\n  -----------------");
       printf("\n  Nb. Segmentations : [%d]",nbsegs); 
       printf("\n  Nb. Classes : [%d]",nbclass);
       printf("\n  Nb. Iters : [%d]",nbiter);
       printf("\n  Value Hyper : [%.1f][%.1f][%.1f][%.1f]",hyperpar[0],hyperpar[1],hyperpar[2],hyperpar[3]);
       printf("\n     (> NbClasses: [%d-[[%d]]-%d])",nbclassinf,nbclass,nbclasssup);
       printf("\n"); }

 //Boucle
 //------
 for(k=0;k<nbsegs;k++)
    {
     seed=10*k+1;

     printf("[%d]",k);
     fflush(stdout);
     
     if (!flag_ssilent)   printf("\n [%d] --> ECI (Seed:%d)",k+1,seed);
     nbclass=((int)nbclassinf)+(int)random_int((int)((int)nbclasssup-(int)nbclassinf)+1);
     if (!flag_ssilent) printf("\n [NbClass %d]\n ",nbclass);
     HistoClasses[nbclass]++;

 
     AutomaticEciSimulation(mt,1);
 
     CopySquareMatrix(mat_seg,tabmat_segneb[k],size_image); 

     if ((flag_save)&&(0)) 
        {
         sprintf(Number,"%d",k);
         SaveImagePgm("SegNB",Number,tabmat_segneb[k],width,length);
        }
    
     tabsetreg[k]=new SetRegion(mat_seg,length,width,sizemax);
     (*tabsetreg[k]).flag_ssilent=flag_ssilent;
     EstimateSetRegions(tabsetreg[k]);
     copy_square_mat(mat_regcoul,tabmat_regcoul[k],size_image,TROIS);
    }
}

//----------------------------------------------------------
// AutomaticEciSimulation              
//----------------------------------------------------------                
void Denoise::AutomaticEciSimulation(float** mt,int flag)
{
 int     i,j;
 int     nbpix;

  //Initialisation
  for(i=0;i<length;i++) for(j=0;j<width;j++) matx_tmp1[i][j]=(int)mt[i][j]; 

  nbpix=width*length;
  for(i=0;i<length;i++) for(j=0;j<width;j++)
    tabpts[(j*width)+i][0]=matx_tmp1[i][j];

  //hyperpar=fmatrix_allocate_1d(4);
  for(i=0;i<4;i++)  hyperpar[i]=value_hyperpar;

  flagR_silent=flag;

  //Affiche Options
  //---------------
  if (!flagR_silent)
     { printf("\n  Simulation ECI: ");
       printf("\n  -----------------");
       printf("\n  Nb Classes : [%d]",nbclass);
       printf("\n  Value Hyper : [%.2f][%.2f][%.2f][%.2f]",hyperpar[0],hyperpar[1],hyperpar[2],hyperpar[3]);
       printf("\n  Value Seed : [%d]",seed); 
       printf("\n  Seulement <Simulations [%d]>",simulations);      
       printf("\n"); }

  //-----------------------------------
  //-------- Algorithme Kmoy ----------
  //-----------------------------------
  km=new kmean(nbclass,tabpts,nbpix,1);
  (*km).FlagKm_silent=flag;
  (*km).flag_ssilent=flag_ssilent;
  (*km).VarClusters=varclusters;
  (*km).Clustering();

  //-----------------------------------
  //-------- Algorithme ECI -----------
  //-----------------------------------
  eci=new Eci(nbclass,length,width,seed);
  (*eci).nbiter=nbiter;
  (*eci).FlagEci_silent=flag;
  (*eci).flag_ssilent=flag_ssilent; 
  (*eci).eci_estimate(matx_tmp1,(*km).MatrixParameterGaussian,hyperpar);
  (*eci).compute_histo(matx_tmp1);
  CopyVect((*eci).ppar,pparseg);
  //if (!flag_ssilent) Visu_screen(0,pparseg);
  for(i=0;i<length;i++) for(j=0;j<width;j++) 
   mat_seg[i][j]=(float)(((*eci).pmatl[i][j]*255)/(nbclass-1));
 
  //Delete Objet
  //------------
  delete km;
  delete eci; 
}

//------------//
//------------//
//------------//
//----------------------------------------------------------
// EstimateSetRegions (SetRegion*)            
//----------------------------------------------------------                
void Denoise::EstimateSetRegions(SetRegion* streg)
{
 //-----------------------------------
 //-------- Algorithme Region --------
 //-----------------------------------
  (*streg).find_regions();
  (*streg).stat_regions();
  Colorization(mat_img,streg);
  (*streg).fill_mat_color();  
  CopySquareMatrix((*streg).mat_coul,mat_regcoul,size_image);
}

//----------------------------------------------------------
// ComputeMatSegMoy    
//----------------------------------------------------------
void Denoise::ComputeMatSegMoy(float** mt,float** mtin)
{
 int i,j,k,l;
 int rw,cl;
 float sum;

 //Calcul MatSegMoy
 //-----------------------------------

 //Remplissage mat_const
  for(l=0;l<nbsegs;l++)                                          
     {
     for(k=0;k<(*tabsetreg[l]).nb_regions;k++)
        {
         sum=0;
         for(i=0;i<(*tabsetreg[l]).pp_region[k]->nb_points;i++)
             { rw=(*tabsetreg[l]).pp_region[k]->pixs[0][i];
               cl=(*tabsetreg[l]).pp_region[k]->pixs[1][i];
               sum+=mtin[rw][cl]; }  

               sum/=(*tabsetreg[l]).pp_region[k]->nb_points;

	 if (1)
         for(i=0;i<(*tabsetreg[l]).pp_region[k]->nb_points;i++)
            { rw=(*tabsetreg[l]).pp_region[k]->pixs[0][i];
              cl=(*tabsetreg[l]).pp_region[k]->pixs[1][i];

              if (rw>length) { rw=length-1; printf("!"); }
              if (cl>width)  { cl=width-1;  printf("!"); }
              if (rw<0)      { rw=0;        printf("!"); }  
              if (cl<0)      { cl=0;        printf("!"); }

              if (sum<0)                               mat_const[l][rw][cl]=0.0;
              if (sum>(NB_GREY_LEVELS-1))              mat_const[l][rw][cl]=(NB_GREY_LEVELS-1);
              if ((sum>=0)&&(sum<=(NB_GREY_LEVELS-1))) mat_const[l][rw][cl]=sum; } 
        }
    }  

  //Moyenne
  for(i=0;i<length;i++) for(j=0;j<width;j++)
     {
      sum=0.0;
      for(l=0;l<nbsegs;l++) sum+=mat_const[l][i][j];             
      sum/=(nbsegs);                                            
  
      mt[i][j]=sum;
     }

  //---------------------------------------
}

//----------------------------------------------------------
// Compute Windows
//----------------------------------------------------------
void Denoise::ComputeWindows()
{
 int i,j,k,l;

 //Windows win_tabmat_segneb
 for(k=0;k<nbsegs;k++) for(i=0;i<size_image;i++) for(j=0;j<size_image;j++)  
   win_tabmat_segneb[i][j+(k*size_image)]=tabmat_segneb[k][i][j];

 //Windows win_tabmat_segregcoul
 for(k=0;k<nbsegs;k++) for(l=0;l<3;l++) for(i=0;i<size_image;i++) for(j=0;j<size_image;j++)  
   win_tabmat_regcoul[l][i][j+(k*size_image)]=tabmat_regcoul[k][l][i][j];
}

//-----------//
//--MESURES--//
//-----------//
//----------------------------------------------------------
// Compute ISNR                
//----------------------------------------------------------     
float Denoise::computeISNR(float** matimg,float** matdeg,float** matrest)
{
 int i,j;
 float num,den;
 float isnr;

 //Boucle
 num=0.0; den=0.0; 
 for(i=0;i<length;i++) for(j=0;j<width;j++) num+=CARRE(matimg[i][j]-matdeg[i][j]);
 for(i=0;i<length;i++) for(j=0;j<width;j++) den+=CARRE(matimg[i][j]-matrest[i][j]);

 //ISNR
 isnr=10*log10(num/den);

 //retour
 return isnr;
}

//----------------------------------------------------------
// MMSE              
//----------------------------------------------------------     
float Denoise::computeMMSE(float** mat1,float** mat2,int sz)
{
 int i,j;
 float sum;
 float mmse;

 //Boucle 
 sum=0.0;
 for(i=0;i<sz;i++) for(j=0;j<sz;j++) 
    sum+=CARRE(mat2[i][j]-mat1[i][j]);

 sum/=(CARRE(sz));

 //MMSE
 mmse=sum;

 //retour
 return mmse;
}

//----------------------------------------------------------
// Compute PSNR                
//----------------------------------------------------------     
float Denoise::computePSNR(float** matimg,float** matrest,int sz)
{
 int i,j;
 float num,den;
 float psnr;

 //Boucle
 num=(sz*sz*CARRE(255.0));

 den=0.0;
 for(i=0;i<sz;i++) for(j=0;j<sz;j++) den+=CARRE(matimg[i][j]-matrest[i][j]);

 //ISNR
 psnr=10*log10(num/den);

 //retour
 return psnr;
}

//----------------------------------------------------------
// Compute SNR                
//----------------------------------------------------------     
float Denoise::computeSNR(float** mt,float vr,int sz)
{
 int i,j;
 float vari,moyi;
 float snr;

 //Image
 vari=0.0;
 moyi=0.0;
 for(i=0;i<sz;i++) for(j=0;j<sz;j++) moyi+=mt[i][j];
 moyi/=(sz*sz);
 for(i=0;i<sz;i++) for(j=0;j<sz;j++) vari+=CARRE(mt[i][j]-moyi);
 vari/=(sz*sz);
 
 //SNR
 snr=10*log10(vari/vr);

 //retour
 return snr;
}

//----------------------------------------------------------
// Compute SNR                
//----------------------------------------------------------     
float Denoise::computeSNR(float** mt,float** mto,int sz)
{
 int i,j;
 float vari,moyi;
 float varb,moyb;
 float snr;

 //Image
 vari=0.0;
 moyi=0.0;
 for(i=0;i<sz;i++) for(j=0;j<sz;j++) moyi+=mto[i][j];
 moyi/=(sz*sz);
 for(i=0;i<sz;i++) for(j=0;j<sz;j++) vari+=CARRE(mto[i][j]-moyi);
 vari/=(sz*sz);

 if (!flag_ssilent)
 printf("\n  > Moyenne Img = [%.0f]       Var. Image = [%.0f]",moyi,vari);

 //Bruit
 varb=0.0;
 moyb=0.0;
 for(i=0;i<sz;i++) for(j=0;j<sz;j++) moyb+=(mto[i][j]-mt[i][j]);
 moyb/=(sz*sz);
 for(i=0;i<sz;i++) for(j=0;j<sz;j++) varb+=CARRE(mto[i][j]-mt[i][j]-moyb);
 varb/=(sz*sz);

 if (!flag_ssilent)
 printf("\n  > Moyenne Bruit = [%.2f]   Var. Bruit = [%.2f]\n",moyb,varb);
 
 //SNR
 snr=10*log10(vari/varb);

 //retour
 return snr;
}


//---------//
//--UTILE--//
//---------//
//----------------------------------------------------------
// Sauvegarde de l'image de nom <name> au format pgm                        
//----------------------------------------------------------                
void Denoise::saveImagePgm(char* name,char* comment,float** mat,int lgth,int wdth)
{
 int i,j;
 char buff[300];
 FILE* fic;

  //--extension--
  strcpy(buff,name);
  strcat(buff,comment);
  strcat(buff,".pgm");

  //--ouverture fichier--
  fic=fopen(buff,"wb");
    if (fic==NULL) 
        { printf("Probleme dans la sauvegarde de %s",buff); 
          exit(-1); }
  printf("\n Sauvegarde de %s au format pgm\n",buff);

  //--sauvegarde de l'entete--
  fprintf(fic,"P5");
  fprintf(fic,"\n# IMG Module");
  fprintf(fic,"\n%d %d",lgth,wdth);
  fprintf(fic,"\n255\n");

  //--enregistrement--
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
	fprintf(fic,"%c",(char)mat[i][j]);
   
  //--fermeture fichier--
   fclose(fic); 
}

//----------------------------------------------------------
// Colorization      
//----------------------------------------------------------                
void Denoise::Colorization(float** mt,SetRegion* streg)
{
 int i,k;
 int row,col;
 float sum;

  //Composante Y
  for (k=0;k<(*streg).nb_regions;k++)
    {
     sum=0;
     for (i=0;i<(*streg).pp_region[k]->nb_points;i++)
	  { row=(*streg).pp_region[k]->pixs[0][i];
	    col=(*streg).pp_region[k]->pixs[1][i];
            sum+=mt[row][col]; }  

     sum/=(*streg).pp_region[k]->nb_points;

     if (sum>255) sum=255.0;
     if (sum<0)   sum=0.0;
     (*streg).pp_region[k]->CoulR=(int)((randomize2()*255));  
     (*streg).pp_region[k]->CoulV=(int)((sum/1.5)+(randomize2()*80));
     (*streg).pp_region[k]->CoulB=(int)((sum/1.5)+(randomize2()*80)); 
    }
}

//----------------------------------------------------------
// FillParRestoration     
//----------------------------------------------------------                
void Denoise::FillResRestoration()
{
 char bufftmp[300];

 //Info
 sprintf(bufftmp,"_Nois%.1f",variance_noise);
 strcpy(Par_Result,bufftmp);

 // sprintf(bufftmp,"_M%d",method);
 //  strcat(Par_Result,bufftmp);

 //  sprintf(bufftmp,"_Ns%d",nbsegs);
 //  strcat(Par_Result,bufftmp);

 //  sprintf(bufftmp,"_NbC%d-%d-%d",nbclass,nbclassinf,nbclasssup);
 //  strcat(Par_Result,bufftmp);

 //  sprintf(bufftmp,"_H%.3f",value_hyperpar);
 //  strcat(Par_Result,bufftmp);

 //  sprintf(bufftmp,"_Vc%.3f",varclusters);
 //  strcat(Par_Result,bufftmp);

 //  sprintf(bufftmp,"_x%d",sizemax);
 //  strcat(Par_Result,bufftmp);

 //  sprintf(bufftmp,"_NbIt%d",nbiter);
 //  strcat(Par_Result,bufftmp);
 
 sprintf(bufftmp,"_MSE_%.2f",mmse_restor);
 strcat(Par_Result,bufftmp);
}

//----------------------------------------------------------
// CopyVect
//----------------------------------------------------------
void Denoise::CopyVect(float** parin,float** parout)
{
 int k;

 //Boucle
 for(k=0;k<nbclass;k++) 
    { parout[k][0]=parin[k][0]; parout[k][1]=parin[k][1]; parout[k][2]=parin[k][2]; }
}

//----------------------------------------------------------
// visualisation sur ecran des differents parametres  
//----------------------------------------------------------
void Denoise::Visu_screen(int i,float** ppar)
{
 int k;

 //affichage valeur des parametres
 printf("\n\n %dieme iteration",i);

 for(k=0;k<nbclass;k++) 
 printf("\n prop=%0.3f  moy=%0.1f  var=%0.2f",ppar[k][0],ppar[k][1],ppar[k][2]); 
}

//----------------------------------------------------------
// DiffMat
//----------------------------------------------------------
void Denoise::DiffMat(float** matin,float** mat1,float** mat2,int sz)
{
 int i,j;
 
 //Diff
 for(i=0;i<sz;i++) for(j=0;j<sz;j++) matin[i][j]=(mat1[i][j]-mat2[i][j]);

 //Recalage
 Recal(matin,sz,sz);
}
