//---------------------------------------------------
// module  : Eci.cc                        
// auteur  : Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "Eci.h"

//------------------------------------------------
// METHODES DE LA CLASSE ECI ---------------------
//------------------------------------------------
//----------------------------------------------------------
// constructeur de la classe Eci
//----------------------------------------------------------
 Eci::Eci(int Nbclass,int Length,int Width,int Sd):GibbsSampler(Length,Width,Nbclass)
 {
   //--allocation memoire--- 
   histo=fmatrix_allocate_1d(IMAGE_GREY_LEVEL);
   pparinit=fmatrix_allocate_2d(nbclass,NBPAR);
   ppar=fmatrix_allocate_2d(nbclass,NBPAR);
   pparmoy=fmatrix_allocate_2d(nbclass,NBPAR);
   pparstock=fmatrix_allocate_3d(NBECH,nbclass,NBPAR); 
   img_histo=fmatrix_allocate_2d(LENGTH_HISTO,IMAGE_GREY_LEVEL);
   img_mv=fmatrix_allocate_2d(length,width);

   //Init parametres
   seed=Sd;
   nbiter=ITER;
  
   //InitFlag
   FlagEci_silent=0;
  }

//----------------------------------------------------------
// destructeur  de la classe Eci
//----------------------------------------------------------
 Eci::~Eci()
 { 
  //---libere la memoire---
  free_fmatrix_1d(histo);
  free_fmatrix_2d(pparinit);
  free_fmatrix_2d(ppar);
  free_fmatrix_2d(pparmoy);
  free_fmatrix_3d(pparstock,NBECH);
  free_fmatrix_2d(img_histo); 
  free_fmatrix_2d(img_mv);
 }

//------------------
//------------------
//----------------------------------------------------------
// visualisation sur ecran des differents parametres  
//----------------------------------------------------------
void Eci::visu_screen(int i,float** ppar)
{
 int k;

 //affichage valeur des parametres
 if (!FlagEci_silent) printf("\n\n %dieme iteration",i);

 if (!FlagEci_silent)
 for(k=0;k<nbclass;k++) 
    printf("\n %d> prop=%0.3f  moy=%0.1f  var=%0.2f",k,ppar[k][0],ppar[k][1],ppar[k][2]); 

 
 if (!FlagEci_silent) printf("\n"); 
}

//----------------------------------------------------------
// remplie une matrice avec une segmentation MV                    
//----------------------------------------------------------
void Eci::fill_picture(float** img_mv,int** pmatl)
{
 int i,j;

 //init matrice
 init_label_matrix_MV();
              
 //remplissage de la matrice img_mv au sens du MV
 for(i=0;i<length;i++) for(j=0;j<width;j++)
 img_mv[i][j]=(((float)pmatl[i][j]*(float)GREY_LEVEL)/(float)(nbclass-1)); 
}
 
//---------------------------------------------------
// calcule un critere d'arret a partir du tableau    
// des NBECH dernieres estimations et la valeur      
// courante des parametres                           
//---------------------------------------------------
float Eci::varpparstock()
{
 int i,j;
 float max,tmp,tmpa,tmpb,minech,maxech;

 max=0.0; 
 
 for(j=0;j<nbclass;j++)
  {
   minech=1000000; maxech=0.0;
     { 
      for(i=0;i<NBECH;i++)
          { 
            tmp=pparstock[i][j][0];  
            if (tmp>maxech) maxech=tmp;
            if (tmp<minech) minech=tmp;
          }
            tmpa=(float)fabs(1-((ppar[j][0])/maxech));
            tmpb=(float)fabs(1-((ppar[j][0])/minech));

            if (tmpa>tmpb) max=tmpa; else max=tmpb;          
     }
  }
 return (100*max);
}

//---------------------------------------------------
// donne la moyenne des estimations enregistres dans 
// le tableau pparstock                              
//---------------------------------------------------
void Eci::moypar()
{
 int i,j,k;
 float moy;
 
 for(j=0;j<nbclass;j++) for(k=0;k<NBPAR;k++)
  {
   moy=0.0;

   for(i=0;i<NBECH;i++) moy+=pparstock[i][j][k];

   moy/=(float)NBECH;
   pparmoy[j][k]=moy;
  }
}

//---------------------------------------------------
// rempli le tableau pparstock des estimations       
//---------------------------------------------------
void Eci::fill_pparstock(int iter)
{
 int k,j,i;

 for(j=0;j<nbclass;j++) for(k=0;k<NBPAR;k++)
   { i=(iter%NBECH); pparstock[i][j][k]=ppar[j][k]; }
}

//------------------
//----------------------------------------------------------
// estimation des differents parametres du melange       
// initiale           
//----------------------------------------------------------
void Eci::init_parameter(int** img,float** ppar_init)
{
 fill_init_matp(img,ppar_init);
 init_label_matrix_MV();
}
//----------------------------------------------------------
// estimation des differents parametres du melange       
// initiale           
//----------------------------------------------------------
void Eci::init_parameter(int** img,float** ppar_init,float* tresh)
{
 int i,j;
 int pix;

 //Initialisation -> treshold
 for(i=0;i<length;i++) for(j=0;j<width;j++)
 { pix=(int)img[i][j]; 
   if (pix<tresh[0])  pmatl[i][j]=0;
   if ((pix>=tresh[0])&&(pix<tresh[1])) pmatl[i][j]=1;
   if (pix>=tresh[1]) pmatl[i][j]=2; }

 //Recherche du min
 ppar_init[0][1]=-1.0;

 //Calcul parametres
 compute_parameter(ppar_init,img);
}

//----------------------------------------------------------
// estimation des differents parametres du melange       
// (estimateur du maximum de vraissemblance) pour le     
// cas loi de rayleigh et loi normale.                   
//----------------------------------------------------------
void Eci::compute_parameter(float** ppar,int** img)
{
  int i,j,k;
  float moy,var,nb;
  float pix;

    //-----iteration-----
    //-------------------
    for(k=0;k<nbclass;k++)
      {
        moy=0.0; var=0.0; nb=0.0; 
         
           //--> moy  
           for(i=0;i<length;i++) for(j=0;j<width;j++)
             if (pmatl[i][j]==k)
                { pix=img[i][j]; 
		  nb++; moy+=pix; }  

               if (nb!=0) moy/=nb;

	    //--> var
           for(i=0;i<length;i++) for(j=0;j<width;j++)
              if (pmatl[i][j]==k)
		 { pix=img[i][j];
		   var+=CARRE(pix-moy); } 

      //--> variance 
      if (nb) var/=nb;
      else var=(float)1; 
     
      //--> proportion 
      nb/=(width*length);  ppar[k][0]=nb;

      //--> moyenne 
      ppar[k][1]=moy;   

      //enregistrement 
      if (var!=0)  ppar[k][2]=var; 
     else ppar[k][2]=(float)1;        
    }
} 

//----------------//
//----------------//
//  ECI procedure //
//----------------//
//----------------//
//---------------------------------------------------------
//  Estimation eci                  
//---------------------------------------------------------
void Eci::eci_estimate(int** img,float** ppar_init,float* hyperpar)
{
  int i,j;
  float stopcond;

  //Init Seed
  sgenrand(seed);

  //ppar_init --> ppar 
  for(i=0;i<nbclass;i++) for(j=0;j<NBPAR;j++) 
  ppar[i][j]=ppar_init[i][j];  
                                                                                 
  //init chp d etiq.au sens du MV                             
  fill_init_matp(img,ppar);  
  init_label_matrix_MV(); 
  visu_screen(0,ppar);                                          

  //____ITERATION__________
  //-----------------------
  for(i=1;i<nbiter;i++)
   {
    //simule un chps et calcule PHI
    iteration(img,ppar,hyperpar);
    compute_parameter(ppar,img); 
    fill_pparstock(i);
    visu_screen(i,ppar); 
    if (!flag_ssilent) printf("[%d]",i);  
    fflush(stdout);     
                                                                                                        
    //critere d arret
    if (i>NBECH)
       {
        stopcond=varpparstock();
        //printf("[%.2f]",stopcond);                                   
        if (stopcond<STOPCOND) 
             { 
              moypar(); 
              visu_screen(0,pparmoy);
              break; 
             }                         
       }  
   } 
  //------------------------
  //------------------------

   // moyen. des dernieres estimations 
   if (ITER>NBECH) moypar();

   // ppar  --> ppar_init 
   for(i=0;i<nbclass;i++) for(j=0;j<NBPAR;j++) 
       {
        if (ITER>NBECH) 
           { ppar_init[i][j]=pparmoy[i][j]; 
             ppar[i][j]=pparmoy[i][j]; }

        else  ppar_init[i][j]=ppar[i][j];
       }

   // visu finale
   visu_screen(0,ppar_init);
}

//-------------//
//-------------//
// HISTOGRAMME //
//-------------//
//-------------//
//----------------------------------------------------------
// calcule l'histogramme de l'image bmp                  
//----------------------------------------------------------
void Eci::compute_histo(int** img)
 {
  histo_picture(img);
  fill_img_histo();
  fill_distribution_mixture();
 }

//----------------------------------------------------------
// calcule l'histogramme d'une image, resultats dans  
// le tableau de flottant histo                        
//----------------------------------------------------------
void Eci::histo_picture(int** img)
 {
 int i,j;

  //initialisation
  for(i=0;i<IMAGE_GREY_LEVEL;i++) histo[i]=0.0;

  //calcul de l'histogramme
     for(i=0;i<length;i++)
       for(j=0;j<width;j++) 
	     if ((img[i][j]>0)&&(img[i][j]<IMAGE_GREY_LEVEL))  
		    histo[(int)img[i][j]]++; 
		 
  //verification
   if (0)
    for(i=1;i<(IMAGE_GREY_LEVEL-1);i++)
	 if (histo[i]>1.7*(histo[i-1]+histo[i+1])/2)
		 histo[i]=(histo[i-1]+histo[i+1])/2;
 }

//----------------------------------------------------------
// rempli img_histo de l'histogramme de l'image                        
//----------------------------------------------------------                
void Eci::fill_img_histo()
 {
  int i,j;

  //init
  float max=0.0;

  //on rempli img_histo de blanc
  for(i=1;i<(LENGTH_HISTO-1);i++) for(j=1;j<(IMAGE_GREY_LEVEL-1);j++)
   img_histo[i][j]=WHITE_PIXEL;

  //recalage
  for(max=0.0,i=0;i<IMAGE_GREY_LEVEL;i++)
   if (histo[i]>max) max=histo[i];

  for(i=0;i<IMAGE_GREY_LEVEL;i++) 
	 histo[i]=(float)(histo[i]/max)*(float)(LENGTH_HISTO-1); 

  //facteur de normalisation
  normhisto=((length*width)*(LENGTH_HISTO-1))/max;

  //affichage
  for(i=0;i<IMAGE_GREY_LEVEL;i++) for(j=0;j<histo[i];j++) 
    img_histo[IMAGE_GREY_LEVEL-1-j][i]=BLACK_PIXEL;
}

//----------------------------------------------------------
// rempli img_histo du melange de distributions                       
//----------------------------------------------------------                
void Eci::fill_distribution_mixture()
 {
  int k;
  float j,tmp;

  //dessine les lois
  for(k=0;k<nbclass;k++)
    for(j=0.0;j<IMAGE_GREY_LEVEL;j=j+(float)0.01)
      {
      tmp=funcgauss(j,ppar[k][0],ppar[k][1],ppar[k][2]);

       if (tmp<=0) tmp=0.0;
       tmp=(tmp*normhisto); 

       if (tmp>(LENGTH_HISTO-1)) tmp=(LENGTH_HISTO-1);

       img_histo[(int)((LENGTH_HISTO-1)-tmp)][(int)j]=COLOR_COURBE;
     }
 }

//----------------------------------------------------------
// fonction gaussienne
//----------------------------------------------------------
float Eci::funcgauss(float pix,float coef,float moy,float var)
{
return (float)((coef)*(1/(sqrt(2*PI*var)))*exp(-1*(CARRE(pix-moy))/(2*var)));
}



