//---------------------------------------------------
// module  : f_GibbsSampler.cc                         
// auteur  : Max Mignotte                           
// date    :                                              
// langage : C++                                      
// labo    :                                   
//---------------------------------------------------

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "GibbsSampler.h"

//------------------------------------------------
// METHODES DE LA CLASSE GibbsSampler -----------
//------------------------------------------------
//----------------------------------------------------------
// constructeur de la classe GibbsSampler
//----------------------------------------------------------
GibbsSampler::GibbsSampler(int Length,int Width,int Nbclass)
{
 //initialisation
 length=Length; 
 width=Width;  
 nbclass=Nbclass;  

 //allocation memoire
 pmatp=fmatrix_allocate_3d(nbclass,length,width);
 pmatl=imatrix_allocate_2d(length,width);
 pmatl_init=imatrix_allocate_2d(length,width);

 //Initialisation
 Nrj_Prior=0.0;
 Nrj_Likelihood=0.0;

 //InitFlag
 FlagGS_silent=0;
}

//----------------------------------------------------------
// constructeur par recopie de la classe GibbsSampler
//----------------------------------------------------------
GibbsSampler::GibbsSampler(GibbsSampler& gs)
{
 int i,j,k;

 //initialisation
 length=gs.length; 
 width=gs.width;  
 nbclass=gs.nbclass;  

 //allocation memoire
 pmatp=fmatrix_allocate_3d(nbclass,length,width);
 pmatl=imatrix_allocate_2d(length,width);
 pmatl_init=imatrix_allocate_2d(length,width);

 //remplissage
 for(i=0;i<length;i++) for(j=0;j<width;j++)
          pmatl_init[i][j]=gs.pmatl_init[i][j];

 for(j=0;j<width;j++) for(i=0;j<width;j++) for(k=0;k<nbclass;k++)
          pmatp[k][i][j]=gs.pmatp[k][i][j];
 }


//----------------------------------------------------------
// destructeur  de la classe GibbsSampler
//----------------------------------------------------------
GibbsSampler::~GibbsSampler()
{ 
 //liberation memoire
 free_fmatrix_3d(pmatp,nbclass); 
 free_imatrix_2d(pmatl); 
 free_imatrix_2d(pmatl_init);
}

//----------------------------------------------------------
// retourne la probabilite d'un pixel y d'appartenir
// a la classe correspondante aux parametres
//      coef, moy et var d une gaussienne
//----------------------------------------------------------
float GibbsSampler::funcgauss(short pix,float coef,float moy,float var)
{
return (float)((coef)*(1/(sqrt(2*PI*var)))*exp(-1*(CARRE((float)pix-moy))/(2*var)));
}

//----------------------------------------------------------
// retourne un nombre aleatoire entre zero et un
//----------------------------------------------------------
float GibbsSampler::randomize(void)
{ 
return ((float)rand()/RAND_MAX); 
}

//----------------------------------------------------------
// utilise dans l'echantilloneur de gibbs
// on selectionne l' une des classes ek
// selon les probabilites d'appartenace
// Pxi/yi(ek/yi)
//----------------------------------------------------------
void GibbsSampler::choose_gibbs_label(int row,int col)
{
 float rd,tmp;
 int k;

  tmp=0.0;

  //choisi l etiquette 
  for(rd=randomize2(),k=0;k<nbclass;k++)    
       { 
         tmp+=pmatp[k][row][col];
         if (tmp>rd) { pmatl[row][col]=k; break; }
       }
}

//----------------------------------------------------------
// renvoie un flottant correspondant a l'energie
// d'attache au donnee:fonction des par. des lois,
// de l'obs., de la loi considere,de la pos. et de
// l'etiq. du site courant
//----------------------------------------------------------
float GibbsSampler::data_connexion_nrj(float** ppar,int**  img,int row,int col,int classe)
{
 float nrj=0.0;
 short pix;

 //recuperation du pixel
 pix=(short)img[row][col];

 //calcul nrj
 nrj=(float)-log(funcgauss(pix,1,ppar[classe][1],ppar[classe][2]));

 return nrj;
}

//----------------------------------------------------------
// renvoie un flottant correspondant a l'energie
// contextuel: fonction du chps d'etiq. prec.,des
// parametres de la distribution de gibbs, de la pos
// et de l'etiq du site courant
//----------------------------------------------------------
float GibbsSampler::contextual_nrj(float *hyperpar,int row,int col,int classe)
{
 float nrj;

 //init
 nrj=0;

 //cliques horizontales
  if (col>0)           if (classe!=pmatl[row][col-1]) nrj+=hyperpar[0];
  if (col<(width-1))   if (classe!=pmatl[row][col+1]) nrj+=hyperpar[0];

 //cliques verticales
 if (row>0)            if (classe!=pmatl[row-1][col]) nrj+=hyperpar[1];
 if (row<(length-1))   if (classe!=pmatl[row+1][col]) nrj+=hyperpar[1];

 //cliques horizontale droite
 if ((col<(width-1))&&(row<(length-1))) if (classe!=pmatl[row+1][col+1]) nrj+=hyperpar[2];
 if ((col>0)&&(row>0))                  if (classe!=pmatl[row-1][col-1]) nrj+=hyperpar[2];

 //cliques horizontale gauches
  if ((col>0)&&(row<(length-1)))    if (classe!=pmatl[row+1][col-1]) nrj+=hyperpar[3];
  if ((col<(width-1))&&(row>0))     if (classe!=pmatl[row-1][col+1]) nrj+=hyperpar[3];

 return nrj;
}

//----------------------------------------------------------
// renvoie un flottant correspondant a l'energie
// contextuel: fonction du chps d'etiq. prec.,des
// parametres de la distribution de gibbs, de la pos
// et de l'etiq du site courant
//----------------------------------------------------------
float GibbsSampler::contextual_nrj2(float *hyperpar,int row,int col,int classe)
{
 float nrj;

 //init
 nrj=0;

 //cliques horizontales
  if (col>0)           if (classe!=pmatl[row][col-1]) nrj+=sqrt(2)*hyperpar[0];
  if (col<(width-1))   if (classe!=pmatl[row][col+1]) nrj+=sqrt(2)*hyperpar[0];

 //cliques verticales
 if (row>0)            if (classe!=pmatl[row-1][col]) nrj+=sqrt(2)*hyperpar[1];
 if (row<(length-1))   if (classe!=pmatl[row+1][col]) nrj+=sqrt(2)*hyperpar[1];

 //cliques horizontale droite
 if ((col<(width-1))&&(row<(length-1))) if (classe!=pmatl[row+1][col+1]) nrj+=(1/sqrt(2))*hyperpar[2];
 if ((col>0)&&(row>0))                  if (classe!=pmatl[row-1][col-1]) nrj+=(1/sqrt(2))*hyperpar[2];

 //cliques horizontale gauches
  if ((col>0)&&(row<(length-1)))    if (classe!=pmatl[row+1][col-1]) nrj+=(1/sqrt(2))*hyperpar[3];
  if ((col<(width-1))&&(row>0))     if (classe!=pmatl[row-1][col+1]) nrj+=(1/sqrt(2))*hyperpar[3];

 return nrj;
}

//----------------------------------------------------------
// une iteration de l'echantilloneur de gibbs
// Temperature = 1
//----------------------------------------------------------
void GibbsSampler::iteration(int** img,float** ppar,float* hyperpar)
{
 int i,j,k;
 float tmp,sum;

 //iteration   
   for(i=0;i<length;i++) for(j=0;j<width;j++)
       {
        sum=0;

         for(k=0;k<nbclass;k++)
           {
            tmp=(float)exp((-(data_connexion_nrj(ppar,img,i,j,k)+contextual_nrj(hyperpar,i,j,k))));
            pmatp[k][i][j]=tmp;  
            sum+=tmp;
           }

   for(k=0;k<nbclass;k++)  pmatp[k][i][j]/=sum;
   choose_gibbs_label(i,j);
   }
}

//----------------------------------------------------------
// une iteration de l'echantilloneur de gibbs
// Temperature = temp
//----------------------------------------------------------
void GibbsSampler::iteration(int** img,float** ppar,float* hyperpar,float temperature)
{
 int i,j,k;
 float tmp,sum;

 //iteration   
   for(i=0;i<length;i++) for(j=0;j<width;j++)
       {
        sum=0;

         for(k=0;k<nbclass;k++)
           {
            tmp=(float)exp(-(1.0/(float)temperature)*(data_connexion_nrj(ppar,img,i,j,k)+contextual_nrj(hyperpar,i,j,k)));
            pmatp[k][i][j]=tmp;  
            sum+=tmp;
           }

   for(k=0;k<nbclass;k++)  pmatp[k][i][j]/=sum;
   choose_gibbs_label(i,j);
   }
}


//----------------------------------------------------------
// remplie la matrice 4d contenant la probabilite       
// d'appartenance de chaque pixel a chacune des classes 
//----------------------------------------------------------
void GibbsSampler::fill_init_matp(int** img,float** ppar)
{
  int i,j,k;
  short pix;
  float tmp,sum;

  tmp=0.0;

  //--iteration
    for(i=0;i<length;i++)
       for(j=0;j<width;j++)
        {
          pix=(short)img[i][j];
          sum=0.0;
            for(k=0;k<nbclass;k++)
             {
              tmp=funcgauss(pix,1,ppar[k][1],ppar[k][2]);
        
              if (tmp<=0) tmp=0.0;
              pmatp[k][i][j]=tmp; sum+=tmp; 
             }

           for(k=0;k<nbclass;k++)  pmatp[k][i][j]/=sum; 
         }
}

//----------------------------------------------------------
// on rempli le champ d'etiquette initial par sa  
// valeur au sens du MV    
//----------------------------------------------------------
void GibbsSampler::init_label_matrix_MV()
{
 int i,j,k;
 float tmp,pmax;

 tmp=0.0;

   //-- Valeur au sens du MV
   for(i=0;i<length;i++) for(j=0;j<width;j++)
       {
        pmax=0.0;
        for(k=0;k<nbclass;k++)
          { tmp=pmatp[k][i][j];
             if (tmp>pmax) { pmatl[i][j]=k; pmax=tmp;  }
          }
       }
}

//----------------------------------------------------------
// copie une matrice int** dans une autre
//----------------------------------------------------------
void GibbsSampler::copy_mat(int** pmatl1,int** pmatl2)
{
 int i,j;
 int pix;

    for(i=0;i<length;i++) for(j=0;j<width;j++)
         { 
           pix=pmatl1[i][j];
           pmatl2[i][j]=pmatl1[i][j]; 
         }
}

//----------------------------------------------------------
// calcule le nombre de differences a chaque iteration
//----------------------------------------------------------
int GibbsSampler::diff_pmatl(int** pmatl1,int** pmatl2)
{
 int i,j;
 int chg=0;

    for(i=0;i<length;i++) for(j=0;j<width;j++)
        { 
         if ((pmatl1[i][j])!=(pmatl2[i][j])) chg++; 
             pmatl2[i][j]=pmatl1[i][j];
        }
 return chg;
}

//----------------------------------------------------------
// calcule l'hamiltonien d'une realisation de x
//----------------------------------------------------------
float GibbsSampler::compute_nrj(int** img,float** ppar,float* hyperpar)
{
 int i,j;
 int lab;
 float tmp,sum=0;
 
  //calcul de l hamiltonien de l image
    for(i=1;i<(length-1);i++) for(j=1;j<(width-1);j++)
        {
         lab=pmatl[i][j];
         tmp=data_connexion_nrj(ppar,img,i,j,lab)+contextual_nrj(hyperpar,i,j,lab);
         sum+=tmp;
        }
  return sum;
}


//----------------------------------------------------------
// calcule l'hamiltonien d'une realisation de x
//----------------------------------------------------------
float GibbsSampler::compute_2nrj(int** img,float** ppar,float* hyperpar)
{
 int i,j;
 int lab;
 float tmp,sum=0;

  //Initialisation
  Nrj_Prior=0.0;
  Nrj_Likelihood=0.0;
 
  //calcul de l hamiltonien de l image
    for(i=1;i<(length-1);i++) for(j=1;j<(width-1);j++)
        {
         lab=pmatl[i][j];
         tmp=data_connexion_nrj(ppar,img,i,j,lab)+contextual_nrj(hyperpar,i,j,lab);
         Nrj_Likelihood+=data_connexion_nrj(ppar,img,i,j,lab);
         Nrj_Prior+=contextual_nrj(hyperpar,i,j,lab); 
         sum+=tmp;
        }

    //Affichage
    printf("  [Likelihood:%.0f][Prior:%.0f]\n",Nrj_Likelihood,Nrj_Prior);
  
  //Retour
  return sum;
}






