//------------------------------------------------------
// module  : f_kmean.cc
// auteur  : Mignotte Max
// date    : 
//------------------------------------------------------
//

//------------------------------------------------
// FICHIERS INCLUS -------------------------------
//------------------------------------------------
#include "kmean.h"
#include "FunctionDen.h"

//---------------------------------------------
// CONSTANTES & DEFINITIONS ----------------------
//------------------------------------------------

//------------------------------------------------
// FONCTIONS DE LA K_MEAN ------------------------
//------------------------------------------------
//----------------------------------------------------------
// Constructeur de la classe K_MEAN
//----------------------------------------------------------
 kmean::kmean(int nbclusters,float** tabpoints,int nbpoints,int nbparam)
 {
  int i,j;

  //Recup parametres
  NbClusters=nbclusters;
  NbPoints=nbpoints;
  NbParam=nbparam; 

  //Allocation memoire 
  VectLabel=imatrix_allocate_1d(NbPoints); 
  MatrixPoints=fmatrix_allocate_2d(NbPoints,NbParam);
  MatrixCenters=fmatrix_allocate_2d(NbClusters,NbParam);
  MatrixParameterGaussian=fmatrix_allocate_2d(NbClusters,TROIS);

  //InitFlag
  FlagKm_silent=0;
   
  //Rempli 
  for(i=0;i<nbpoints;i++) for(j=0;j<nbparam;j++)
     MatrixPoints[i][j]=tabpoints[i][j];

  //Visu
  if (0)
     {
      printf("\n\n Tableau A Clusterise");
      printf("\n --------------------");
      for(i=0;i<NbPoints;i++) 
        { 
	 //printf("\n");
         for(j=0;j<NbParam;j++) printf(" [%.2f]",MatrixPoints[i][j]); 
        }
     }

  //Verification
  if (NbClusters>NbPoints)
     { printf("\n\n Erreur plus de centres que de points\n\n"); exit(-1); } 
  }

//----------------------------------------------------------
// Destructeur de la classe K_MEAN
//----------------------------------------------------------
 kmean::~kmean()
 { 
  //---libere la memoire---
  free_imatrix_1d(VectLabel);
  free_fmatrix_2d(MatrixCenters);
  free_fmatrix_2d(MatrixPoints);
  free_fmatrix_2d(MatrixParameterGaussian);
 }

// --------------
//  GESTION
// --------------
//----------------------------------------------------------
// Init les centres                      
//----------------------------------------------------------
void kmean::InitCenters()
{
 int j,k;

 //Boucle
 for(k=0;k<NbClusters;k++) for(j=0;j<NbParam;j++)
   {
    MatrixCenters[k][j]=k*MAXGREYVALUE/NbClusters;
   } 
}

//----------------------------------------------------------
// Init les centres                      
//----------------------------------------------------------
void kmean::InitCenters2()
{
 int j,k;
 float aleat;

 //Affiche
 if (!flag_ssilent)
 printf("\n >> Randomized Kmean Centers [%.3f]",VarClusters);

 //Boucle
 for(k=0;k<NbClusters;k++) for(j=0;j<NbParam;j++)
   {
    aleat=VarClusters*((1-randomize())/2)*(MAXGREYVALUE/2*NbClusters);
    MatrixCenters[k][j]=(k*MAXGREYVALUE/NbClusters)+aleat;
   } 
}

//----------------------------------------------------------
// Une iteration de clustering                         
//----------------------------------------------------------
int kmean::Cluster()
{
 int i,j,k; 
 int flag;
 int classe;
 float dist,distmin;

 //Init
 flag=0; classe=0;

 //Boucle
 for(k=0;k<NbPoints;k++)
    {
     distmin=BIGNUMBER;
 
      for(i=0;i<NbClusters;i++)
  	 { 
          dist=0.0;
	  for(j=0;j<NbParam;j++) dist+=CARRE(MatrixCenters[i][j]-MatrixPoints[k][j]);
   
	  if (dist<distmin) { distmin=dist; classe=i; } 
         }
      if (VectLabel[k]!=classe) { VectLabel[k]=classe; flag=1; } 
    }

 //Return
 return flag;
}


//----------------------------------------------------------
// Calcule le centre des differents clusters        
//----------------------------------------------------------
void kmean::ComputeMean()
{
 int i,j,k;
 int compteur;
 float moytmp[NbParam];
 
 //Boucle
 for(k=0;k<NbClusters;k++)
    {
     compteur=0;
     for(j=0;j<NbParam;j++)  moytmp[j]=0.0;

     for(i=0;i<NbPoints;i++) 
        {
	 if (VectLabel[i]==k) 
	    {
	     compteur++;
             for(j=0;j<NbParam;j++) moytmp[j]+=MatrixPoints[i][j];
            }
        }
     for(j=0;j<NbParam;j++) moytmp[j]/=(float)compteur;
     for(j=0;j<NbParam;j++) MatrixCenters[k][j]=moytmp[j];
    }
}

//----------------------------------------------------------
// Visu les centres                      
//----------------------------------------------------------
void kmean::VisuCenters()
{
 int j,k;

 //Boucle
 for(k=0;k<NbClusters;k++) 
    { printf("\n ");
      for(j=0;j<NbParam;j++) printf("[%.2f]",MatrixCenters[k][j]); }   
}

//----------------------------------------------------------
// Visu les classes                      
//----------------------------------------------------------
void kmean::VisuClasses()
{
 int i,j;

 if (!FlagKm_silent)
    {
     printf("\n\n\n Tableau Clusterise");
     printf("\n --------------------");
    }

  for(i=0;i<NbPoints;i++) 
     { 
       //printf("\n");
       for(j=0;j<NbParam;j++) printf(" [%.2f]",MatrixPoints[i][j]); 
       printf(" <%d>",VectLabel[i]); }
  printf("\n\n\n");
}

//----------------------------------------------------------
// Calcul parametre Gausiens                    
//----------------------------------------------------------
void kmean::ComputeParameterGaussian()
{
  int i,k;
  float moy,var,nb;
  float pix;

  if (!FlagKm_silent)
     {
     printf("\n\n\n Calcul Parametre Gaussien");
     printf("\n -------------------------");
     }

    //-----iteration-----
    //-------------------
    for(k=0;k<NbClusters;k++)
      {
        moy=0.0; var=0.0; nb=0.0; 
         
           //--> moy  
           for(i=0;i<NbPoints;i++) 
              if (VectLabel[i]==k)
                 { pix=MatrixPoints[i][0]; 
		   nb++; moy+=pix; }  

	       if (nb!=0) moy/=nb;

	    //--> var
           for(i=0;i<NbPoints;i++) 
              if (VectLabel[i]==k)
	         { pix=MatrixPoints[i][0];
		   var+=CARRE(pix-moy); } 

      //--> variance 
      if (nb) var/=nb;
      else var=1.0;
      MatrixParameterGaussian[k][2]=var;

      //--> moyenne 
      MatrixParameterGaussian[k][1]=moy; 
     
      //--> proportion 
      nb/=(NbPoints);  
      MatrixParameterGaussian[k][0]=nb; 
      }    
}

//----------------------------------------------------------
// Print parametre Gausiens                    
//----------------------------------------------------------
void kmean::PrintParameterGaussian()
{
  int k;

 //affichage valeur des parametres
 printf("\n\n Parametre Gaussiens : ");

 if (!flag_ssilent)
 for(k=0;k<NbClusters;k++)
    { 
     printf("\n prop=%.3f  moy=%.1f",MatrixParameterGaussian[k][0],MatrixParameterGaussian[k][1]); 
     printf(" var=%.2f",MatrixParameterGaussian[k][2]); 
    }

 if (!flag_ssilent) printf("\n");
}

//--------------------------------
// FONCTION CLIENTS --------------
//--------------------------------
//----------------------------------------------------------
//  Clustering                     
//----------------------------------------------------------
void kmean::Clustering()
{
  int k;

  //Initialisation
  if (!VarClusters) InitCenters();
  if (VarClusters)  InitCenters2();
  if (!FlagKm_silent) printf("\n\n  K-Moy >  ");

  //Boucle
  for(k=0;;k++)
     { 
       if (!(k%20)) if (!flag_ssilent) if (!FlagKm_silent) printf("\n  ");   
       if (!flag_ssilent) if (!FlagKm_silent) printf("[%d]",k);
       if (!Cluster()) break;
       ComputeMean();
       //VisuCenters(); 
     }  

  //Estimation parametre
  //VisuClasses();
  ComputeParameterGaussian();
  if (!FlagKm_silent) PrintParameterGaussian();
} 










